from __future__ import annotations

import networkx as nx  # type: ignore
from networkx import DiGraph

from pyhcl2.nodes import Block, Module
from pyhcl2.tracker import resolve_variable_references


def get_blocks_by_generation(
    module: Module, block_type: str, reverse: bool = False
) -> list[list[Block]]:
    blocks = module.get_blocks(block_type)
    generations = _topological_generations(blocks)
    if reverse:
        generations.reverse()
    return generations


def _topological_generations(blocks: list[Block]) -> list[list[Block]]:
    blocks_by_key = {block.key(): block for block in blocks}

    graph: DiGraph = nx.DiGraph()

    for key, block_under_test in blocks_by_key.items():
        graph.add_node(key)
        variable_references = resolve_variable_references(block_under_test)

        for dirty_child in variable_references:
            if dirty_child in blocks_by_key:
                graph.add_edge(dirty_child, key)

    if not nx.is_directed_acyclic_graph(graph):
        raise ValueError("Graph is not a DAG")

    generations = list(nx.topological_generations(graph))

    return [[blocks_by_key[k] for k in generation] for generation in generations]



from __future__ import annotations

import re
import sys
from typing import TypeVar, cast

from lark import Discard, Token, Transformer, v_args
from lark.tree import Meta
from lark.visitors import _DiscardType
from pyagnostics.spans import SourceSpan

from pyhcl2.nodes import (
    ArrayExpression,
    Attribute,
    AttrSplat,
    BinaryExpression,
    BinaryOperator,
    Block,
    Conditional,
    Expression,
    ForObjectExpression,
    ForTupleExpression,
    FunctionCall,
    GetAttr,
    GetAttrKey,
    GetIndex,
    GetIndexKey,
    Identifier,
    IndexSplat,
    Literal,
    Node,
    ObjectExpression,
    Parenthesis,
    Stmt,
    UnaryExpression,
    UnaryOperator,
    VarArgsMarker,
)
from pyhcl2.values import Boolean, Float, Integer, Null, String

HEREDOC_PATTERN = re.compile(r"<<([a-zA-Z][a-zA-Z0-9._-]+)\n((.|\n)*?)\n\s*\1", re.S)
HEREDOC_TRIM_PATTERN = re.compile(
    r"<<-([a-zA-Z][a-zA-Z0-9._-]+)\n((.|\n)*?)\n\s*\1", re.S
)
T = TypeVar("T")


# noinspection PyMethodMayBeStatic
class ToAstTransformer(Transformer):
    @v_args(inline=True)
    def __binary_op(self, op: Token) -> BinaryOperator:
        assert op.start_pos is not None
        assert op.end_pos is not None
        return BinaryOperator(type=op.value, span=SourceSpan(op.start_pos, op.end_pos))

    add_op = __binary_op
    mul_op = __binary_op
    comp_op = __binary_op
    eq_op = __binary_op
    or_op = __binary_op
    and_op = __binary_op

    @v_args(inline=True)
    def __binary_expression(
        self, left: Expression, op: BinaryOperator, right: Expression
    ) -> BinaryExpression:
        return BinaryExpression(
            op,
            left,
            right,
            span=SourceSpan(left.span.start, right.span.end),
        )

    term = __binary_expression
    add_expr = __binary_expression
    compare = __binary_expression
    equality = __binary_expression
    and_test = __binary_expression
    or_test = __binary_expression

    @v_args(inline=True)
    def __unary_op(self, op: Token) -> UnaryOperator:
        assert op.start_pos is not None
        assert op.end_pos is not None
        return UnaryOperator(type=op.value, span=SourceSpan(op.start_pos, op.end_pos))

    not_op = __unary_op
    neg_op = __unary_op

    @v_args(inline=True)
    def __unary_expression(
        self, op: UnaryOperator, expr: Expression
    ) -> UnaryExpression:
        return UnaryExpression(op, expr, span=SourceSpan(op.span.start, expr.span.end))

    not_test = __unary_expression
    neg = __unary_expression

    @v_args(inline=True)
    def conditional(
        self, condition: Expression, then_expr: Expression, else_expr: Expression
    ) -> Conditional:
        return Conditional(
            condition,
            then_expr,
            else_expr,
            span=SourceSpan(condition.span.start, else_expr.span.end),
        )

    @v_args(inline=True, meta=True)
    def get_attr(self, meta: Meta, identifier: Identifier) -> GetAttrKey:
        return GetAttrKey(identifier, span=SourceSpan(meta.start_pos, meta.end_pos))

    @v_args(inline=True)
    def get_attr_expr_term(self, on: Expression, key: GetAttrKey) -> GetAttr:
        return GetAttr(on, key, span=SourceSpan(on.span.start, key.span.end))

    @v_args(meta=True)
    def float_lit(self, meta: Meta, args: list[Token]) -> Literal:
        return Literal(
            Float(float("".join([str(arg) for arg in args]))),
            span=SourceSpan(meta.start_pos, meta.end_pos),
        )

    @v_args(meta=True, inline=True)
    def null_lit(self, meta: Meta) -> Literal:
        span = SourceSpan(meta.start_pos, meta.end_pos)
        return Literal(Null(span=span), span=span)

    @v_args(meta=True)
    def int_lit(self, meta: Meta, args: list[Token]) -> Literal:
        span = SourceSpan(meta.start_pos, meta.end_pos)
        return Literal(
            Integer(int("".join([str(arg) for arg in args])), span=span),
            span=span,
        )

    @v_args(inline=True)
    def expr_term(self, expr: Expression) -> Expression:
        return expr

    @v_args(inline=True)
    def bool_lit(self, token: Token) -> Literal:
        assert token.start_pos is not None
        assert token.end_pos is not None
        span = SourceSpan(token.start_pos, token.end_pos)
        match token.value.lower():
            case "true":
                return Literal(
                    Boolean(True, span=span),
                    span=span,
                )
            case "false":
                return Literal(
                    Boolean(False, span=span),
                    span=span,
                )
        raise ValueError(f"Invalid boolean value: {token.value}")

    @v_args(inline=True)
    def string_lit(self, token: Token) -> Literal:
        assert token.start_pos is not None
        assert token.end_pos is not None
        span = SourceSpan(token.start_pos, token.end_pos)
        return Literal(
            String(token.value[1:-1], span=span),
            span=span,
        )

    @v_args(inline=True)
    def identifier(self, ident: Token) -> Expression:
        assert ident.start_pos is not None
        assert ident.end_pos is not None
        return Identifier(ident.value, span=SourceSpan(ident.start_pos, ident.end_pos))

    @v_args(inline=True)
    def attribute(self, ident: Identifier, expr: Expression) -> Attribute:
        return Attribute(ident, expr, span=SourceSpan(ident.span.start, expr.span.end))

    def body(self, args: list[Stmt]) -> list[Stmt]:
        return args

    @v_args(meta=True)
    def block(self, meta: Meta, args: list[Identifier | Literal | Stmt]) -> Block:
        type_key = cast(Identifier, args[0])
        labels = cast(list[Identifier | Literal], args[1:-1])
        body = cast(list[Stmt], args[-1])
        return Block(
            type_key,
            labels,
            body,
            span=SourceSpan(meta.start_pos, meta.end_pos),
        )

    @v_args(meta=True)
    def object(
        self, meta: Meta, args: list[tuple[Expression, Expression]]
    ) -> ObjectExpression:
        return ObjectExpression(
            {kv[0]: kv[1] for kv in args}, span=SourceSpan(meta.start_pos, meta.end_pos)
        )

    @v_args(inline=True)
    def object_elem(
        self, key: Expression, value: Expression
    ) -> tuple[Expression, Expression]:
        return key, value

    @v_args(meta=True)
    def array(self, meta: Meta, values: list[Expression]) -> ArrayExpression:
        return ArrayExpression(values, span=SourceSpan(meta.start_pos, meta.end_pos))

    @v_args(meta=True, inline=True)
    def paren_expr(self, meta: Meta, expr: Expression) -> Parenthesis:
        return Parenthesis(expr, span=SourceSpan(meta.start_pos, meta.end_pos))

    @v_args(meta=True, inline=True)
    def function_call(
        self,
        meta: Meta,
        ident: Identifier,
        args: tuple[list[Expression], VarArgsMarker | None] = ([], None),
    ) -> FunctionCall:
        return FunctionCall(
            ident,
            args[0],
            args[1] is not None,
            span=SourceSpan(meta.start_pos, meta.end_pos),
        )

    def arguments(
        self, args: list[Expression]
    ) -> tuple[list[Expression], VarArgsMarker | None]:
        if len(args) > 0 and isinstance(args[-1], VarArgsMarker):
            return args[:-1], cast(VarArgsMarker, args[-1])
        return args, None

    @v_args(meta=True, inline=True)
    def ellipsis(self, meta: Meta) -> VarArgsMarker:
        return VarArgsMarker(span=SourceSpan(meta.start_pos, meta.end_pos))

    @v_args(meta=True, inline=True)
    def index(self, meta: Meta, expr: Expression) -> GetIndexKey:
        return GetIndexKey(expr, span=SourceSpan(meta.start_pos, meta.end_pos))

    @v_args(meta=True, inline=True)
    def index_expr_term(
        self, meta: Meta, on: Expression, index: GetIndexKey
    ) -> GetIndex:
        return GetIndex(on, index, span=SourceSpan(meta.start_pos, meta.end_pos))

    def attr_splat(self, args: list[Node]) -> list[Node]:
        return args

    @v_args(meta=True, inline=True)
    def attr_splat_expr_term(
        self, meta: Meta, on: Expression, keys: list[GetAttrKey]
    ) -> AttrSplat:
        return AttrSplat(on, keys, span=SourceSpan(meta.start_pos, meta.end_pos))

    def full_splat(self, args: list[Node]) -> list[Node]:
        return args

    @v_args(meta=True, inline=True)
    def full_splat_expr_term(
        self, meta: Meta, on: Expression, keys: list[GetAttrKey | GetIndexKey]
    ) -> IndexSplat:
        return IndexSplat(on, keys, span=SourceSpan(meta.start_pos, meta.end_pos))

    def new_line_or_comment(self, _args: list[Node | Token]) -> _DiscardType:
        return Discard

    def new_line_and_or_comma(self, _args: list[Node | Token]) -> _DiscardType:
        return Discard

    def start(self, args: list[Node | Token]) -> Node | Token:
        return args[0]

    start_expr = start
    start_expr_or_stmt = start

    def for_intro(
        self, args: list[Node | Token]
    ) -> tuple[Identifier | None, Identifier, Expression]:
        return (
            cast(Identifier, args[0]) if len(args) == 3 else None,
            cast(Identifier, args[1 if len(args) == 3 else 0]),
            cast(Expression, args[-1]),
        )

    @v_args(inline=True)
    def for_cond(self, condition: Expression) -> Expression:
        return condition

    @v_args(meta=True, inline=True)
    def for_tuple_expr(
        self,
        meta: Meta,
        for_intro: tuple[Identifier | None, Identifier, Expression],
        expression: Expression,
        condition: Expression | None = None,
    ) -> ForTupleExpression:
        key_ident, value_ident, collection = for_intro
        return ForTupleExpression(
            key_ident,
            value_ident,
            collection,
            expression,
            condition,
            span=SourceSpan(meta.start_pos, meta.end_pos),
        )

    @v_args(meta=True, inline=True)
    def for_object_expr(
        self,
        meta: Meta,
        for_intro: tuple[Identifier | None, Identifier, Expression],
        key_expression: Expression,
        value_expression: Expression,
        *args: VarArgsMarker | Expression,
    ) -> ForObjectExpression:
        key_ident, value_ident, collection = for_intro
        grouping_mode = any(isinstance(arg, VarArgsMarker) for arg in args)
        condition: Expression | None = next(
            (arg for arg in args if isinstance(arg, Expression)), None
        )

        return ForObjectExpression(
            key_ident,
            value_ident,
            collection,
            key_expression,
            value_expression,
            condition,
            grouping_mode,
            span=SourceSpan(meta.start_pos, meta.end_pos),
        )

    @v_args(meta=True)
    def heredoc_template(self, meta: Meta, args: list[Node | Token]) -> Literal:
        match = HEREDOC_PATTERN.match(str(args[0]))
        if not match:
            raise RuntimeError(f"Invalid Heredoc token: {args[0]}")
        return Literal(
            String(match.group(2)),
            span=SourceSpan(meta.start_pos, meta.end_pos),
        )

    @v_args(meta=True)
    def heredoc_template_trim(self, meta: Meta, args: list[Node | Token]) -> Literal:
        match = HEREDOC_TRIM_PATTERN.match(str(args[0]))
        if not match:
            raise RuntimeError(f"Invalid Heredoc token: {args[0]}")

        text = match.group(2)
        lines = text.split("\n")

        # calculate the min number of leading spaces in each line
        min_spaces = sys.maxsize
        for line in lines:
            leading_spaces = len(line) - len(line.lstrip(" "))
            min_spaces = min(min_spaces, leading_spaces)

        # trim off that number of leading spaces from each line
        lines = [line[min_spaces:] for line in lines]

        return Literal(
            String("\n".join(lines)),
            span=SourceSpan(meta.start_pos, meta.end_pos),
        )



from __future__ import annotations

import typing as t

from lark import Lark, Token, UnexpectedCharacters, UnexpectedToken
from pyagnostics.exceptions import DiagnosticError
from pyagnostics.spans import LabeledSpan, SourceSpan

from pyhcl2.nodes import Expression, Module, Node, Stmt
from pyhcl2.transformer import ToAstTransformer


def parse_file(file: t.TextIO) -> Module:
    return parse_module(file.read())


def parse_string(text: str, start: str) -> Node:
    try:
        lark = Lark.open(
            "hcl2.lark",
            parser="lalr",
            start=start,
            cache=True,
            rel_to=__file__,
            propagate_positions=True,
        )
        parse_tree = lark.parse(text)
        ast = ToAstTransformer().transform(parse_tree)

    except UnexpectedCharacters as e:
        assert e.pos_in_stream is not None
        raise DiagnosticError(
            code="pyhcl2::lexer::unexpected_character",
            message="The lexer encountered an unexpected character",
            labels=[
                LabeledSpan(
                    SourceSpan(e.pos_in_stream, e.pos_in_stream + 1),
                    "Unexpected character",
                )
            ],
        ) from None
    except UnexpectedToken as e:
        if e.token.type == "$END":
            raise DiagnosticError(
                code="pyhcl2::parser::unexpected_eof",
                message="The parser expected a token, but the input ended",
                labels=[
                    LabeledSpan(
                        SourceSpan(e.token.start_pos, e.token.end_pos), "Unexpected EOF"
                    )
                ],
            ) from e
        else:
            assert isinstance(e.token, Token)
            assert e.token.start_pos is not None
            assert e.token.end_pos is not None
            if e.token.value == "\n":
                raise DiagnosticError(
                    code="pyhcl2::parser::unexpected_newline",
                    message="The parser encountered an unexpected newline",
                    labels=[
                        LabeledSpan(
                            SourceSpan(e.token.start_pos, e.token.end_pos),
                            "Unexpected newline",
                        )
                    ],
                ) from None
            else:
                raise DiagnosticError(
                    code="pyhcl2::parser::unexpected_token",
                    message="The parser encountered an unexpected token",
                    labels=[
                        LabeledSpan(
                            SourceSpan(e.token.start_pos, e.token.end_pos),
                            "Unexpected token",
                        )
                    ],
                    notes=[
                        f"Got {e.token.value!r} instead",
                    ],
                ) from None

    return t.cast(Node, ast)


def parse_module(text: str) -> Module:
    return Module(
        t.cast(list[Stmt], parse_string(text, start="start")),
        span=SourceSpan(0, len(text)),
    )


def parse_expr(text: str) -> Expression:
    return t.cast(Expression, parse_string(text, start="start_expr"))


def parse_expr_or_stmt(text: str) -> Expression | Stmt:
    return t.cast(Expression | Stmt, parse_string(text, start="start_expr_or_stmt"))



from __future__ import annotations

import dataclasses
from collections.abc import (
    Iterable,
    Iterator,
    Mapping,
    MutableMapping,
    MutableSequence,
    Sequence,
)
from dataclasses import dataclass, field
from os import PathLike
from typing import (
    Never,
    Self,
    overload,
)

from pyagnostics.exceptions import DiagnosticError
from pyagnostics.spans import LabeledSpan, SourceSpan
from rich.console import Console, ConsoleOptions, ConsoleRenderable, RenderResult
from rich.segment import Segment
from rich.text import Span

import pyhcl2.nodes
from pyhcl2.rich_utils import (
    STYLE_KEYWORDS,
    STYLE_NUMBER,
    STYLE_PROPERTY_NAME,
    STYLE_STRING,
    Inline,
)


@dataclass(kw_only=True, frozen=True)
class Value(ConsoleRenderable):
    span: SourceSpan | None = field(compare=False, hash=False, default=None)

    def with_span(self, span: SourceSpan | None) -> Self:
        return dataclasses.replace(self, span=span)

    def rich_highlights(self) -> Iterable[Span]:
        return []

    @staticmethod
    def infer(raw: object) -> Value:
        value: Value
        match raw:
            case None:
                value = Null()
            case bool() as raw:
                value = Boolean(raw)
            case int() as raw:
                value = Integer(raw)
            case float() as raw:
                value = Float(raw)
            case str() as raw:
                value = String(raw)
            case Sequence() as raw:
                value = Array([Value.infer(item) for item in raw])
            case Mapping() as raw:
                value = Object({String(str(k)): Value.infer(v) for k, v in raw.items()})
            case Value() as raw:
                value = raw
            case PathLike() as raw:
                value = String(str(raw))
            case _:
                raise NotImplementedError(
                    f"Could not infer Value for {raw} of type {type(raw)}"
                )

        return value

    def resolve(self) -> Value:
        return self

    def raise_on_unknown(self) -> Value:
        return self

    def raw(self) -> object: ...

    @property
    def type_name(self) -> str:
        return type(self).__name__.lower()

    def __not_equals__(self, other: Value) -> Boolean:
        try:
            return Boolean(not (getattr(self, "__equals__")(other).raw()))
        except AttributeError:
            return Boolean(True)


@dataclass(eq=True, frozen=True)
class Null(Value):
    def __equals__(self, other: Value) -> Boolean:
        return Boolean(isinstance(other, Null))

    def raw(self) -> None:
        return None

    def __rich_console__(
        self, console: Console, options: ConsoleOptions
    ) -> RenderResult:
        yield Segment("null", style=STYLE_KEYWORDS)

    def rich_highlights(self) -> Iterable[Span]:
        if self.span is not None:
            yield self.span.styled(STYLE_KEYWORDS)


@dataclass(eq=True, frozen=True)
class Integer(Value):
    _raw: int

    def raw(self) -> int:
        return self._raw

    def __add__(self, other: Value) -> Value:
        match other:
            case Integer() as other:
                return Integer(self._raw + other.raw())
            case Float() as other:
                return Float(self._raw + other.raw())
            case _:
                raise NotImplementedError()

    def __sub__(self, other: Value) -> Value:
        match other:
            case Integer() as other:
                return Integer(self._raw - other.raw())
            case Float() as other:
                return Float(self._raw - other.raw())
            case _:
                raise NotImplementedError()

    def __mul__(self, other: Value) -> Value:
        match other:
            case Integer() as other:
                return Integer(self._raw * other.raw())
            case Float() as other:
                return Float(self._raw * other.raw())
            case _:
                raise NotImplementedError()

    def __truediv__(self, other: Value) -> Value:
        match other:
            case Integer() as other:
                return Float(self._raw / other.raw())
            case Float() as other:
                return Float(self._raw / other.raw())
            case _:
                raise NotImplementedError()

    def __mod__(self, other: Value) -> Value:
        match other:
            case Integer() as other:
                return Integer(self._raw % other.raw())
            case Float() as other:
                return Float(self._raw % other.raw())
            case _:
                raise NotImplementedError()

    def __equals__(self, other: Value) -> Boolean:
        match other:
            case Integer() as other:
                return Boolean(self._raw == other.raw())
            case Float() as other:
                return Boolean(self._raw == other.raw())
            case _:
                return Boolean(False)

    def __lt__(self, other: Value) -> Boolean:
        match other:
            case Integer() as other:
                return Boolean(self._raw < other.raw())
            case Float() as other:
                return Boolean(self._raw < other.raw())
            case _:
                raise NotImplementedError()

    def __gt__(self, other: Value) -> Boolean:
        match other:
            case Integer() as other:
                return Boolean(self._raw > other.raw())
            case Float() as other:
                return Boolean(self._raw > other.raw())
            case _:
                raise NotImplementedError()

    def __le__(self, other: Value) -> Boolean:
        match other:
            case Integer() as other:
                return Boolean(self._raw <= other.raw())
            case Float() as other:
                return Boolean(self._raw <= other.raw())
            case _:
                raise NotImplementedError()

    def __ge__(self, other: Value) -> Boolean:
        match other:
            case Integer() as other:
                return Boolean(self._raw >= other.raw())
            case Float() as other:
                return Boolean(self._raw >= other.raw())
            case _:
                raise NotImplementedError()

    def __neg__(self) -> Integer:
        return Integer(-self._raw)

    def __rich_console__(
        self, console: Console, options: ConsoleOptions
    ) -> RenderResult:
        yield Segment(str(self._raw), style=STYLE_NUMBER)

    def rich_highlights(self) -> Iterable[Span]:
        if self.span is not None:
            yield self.span.styled(STYLE_NUMBER)


@dataclass(eq=True, frozen=True)
class String(Value):
    _raw: str

    def raw(self) -> str:
        return self._raw

    def __add__(self, other: Value) -> Value:
        match other:
            case String() as other:
                return String(self._raw + other.raw())
            case _:
                raise NotImplementedError()

    def __mul__(self, other: Value) -> Value:
        match other:
            case Integer() as other:
                return String(self._raw * other.raw())
            case _:
                raise NotImplementedError()

    def __equals__(self, other: Value) -> Boolean:
        match other:
            case String() as other:
                return Boolean(self._raw == other.raw())
            case _:
                return Boolean(False)

    def __rich_console__(
        self, console: Console, options: ConsoleOptions
    ) -> RenderResult:
        yield Segment(repr(self._raw), style=STYLE_STRING)

    def rich_highlights(self) -> Iterable[Span]:
        if self.span is not None:
            yield self.span.styled(STYLE_STRING)


@dataclass(eq=True, frozen=True)
class Float(Value):
    _raw: float

    def raw(self) -> float:
        return self._raw

    def __add__(self, other: Value) -> Value:
        match other:
            case Integer() as other:
                return Float(self._raw + other.raw())
            case Float() as other:
                return Float(self._raw + other.raw())
            case _:
                raise NotImplementedError()

    def __sub__(self, other: Value) -> Value:
        match other:
            case Integer() as other:
                return Float(self._raw - other.raw())
            case Float() as other:
                return Float(self._raw - other.raw())
            case _:
                raise NotImplementedError()

    def __mul__(self, other: Value) -> Value:
        match other:
            case Integer() as other:
                return Float(self._raw * other.raw())
            case Float() as other:
                return Float(self._raw * other.raw())
            case _:
                raise NotImplementedError()

    def __truediv__(self, other: Value) -> Value:
        match other:
            case Integer() as other:
                return Float(self._raw / other.raw())
            case Float() as other:
                return Float(self._raw / other.raw())
            case _:
                raise NotImplementedError()

    def __mod__(self, other: Value) -> Value:
        match other:
            case Integer() as other:
                return Float(self._raw % other.raw())
            case Float() as other:
                return Float(self._raw % other.raw())
            case _:
                raise NotImplementedError()

    def __equals__(self, other: Value) -> Boolean:
        match other:
            case Integer() as other:
                return Boolean(self._raw == other.raw())
            case Float() as other:
                return Boolean(self._raw == other.raw())
            case _:
                return Boolean(False)

    def __lt__(self, other: Value) -> Boolean:
        match other:
            case Integer() as other:
                return Boolean(self._raw < other.raw())
            case Float() as other:
                return Boolean(self._raw < other.raw())
            case _:
                raise NotImplementedError()

    def __gt__(self, other: Value) -> Boolean:
        match other:
            case Integer() as other:
                return Boolean(self._raw > other.raw())
            case Float() as other:
                return Boolean(self._raw > other.raw())
            case _:
                raise NotImplementedError()

    def __le__(self, other: Value) -> Boolean:
        match other:
            case Integer() as other:
                return Boolean(self._raw <= other.raw())
            case Float() as other:
                return Boolean(self._raw <= other.raw())
            case _:
                raise NotImplementedError()

    def __ge__(self, other: Value) -> Boolean:
        match other:
            case Integer() as other:
                return Boolean(self._raw >= other.raw())
            case Float() as other:
                return Boolean(self._raw >= other.raw())
            case _:
                raise NotImplementedError()

    def __neg__(self) -> Float:
        return Float(-self._raw)

    def __rich_console__(
        self, console: Console, options: ConsoleOptions
    ) -> RenderResult:
        yield Segment(str(self._raw), style=STYLE_NUMBER)

    def rich_highlights(self) -> Iterable[Span]:
        if self.span is not None:
            yield self.span.styled(STYLE_NUMBER)


@dataclass(eq=True, frozen=True)
class Boolean(Value):
    _raw: bool

    def raw(self) -> bool:
        return self._raw

    def __and__(self, other: Value) -> Boolean:
        match other:
            case Boolean() as other:
                return Boolean(self._raw and other.raw())
            case _:
                raise NotImplementedError()

    def __or__(self, other: Value) -> Boolean:
        match other:
            case Boolean() as other:
                return Boolean(self._raw or other.raw())
            case _:
                raise NotImplementedError()

    def __not__(self) -> Boolean:
        return Boolean(not self._raw)

    def __equals__(self, other: Value) -> Boolean:
        match other:
            case Boolean() as other:
                return Boolean(self._raw == other.raw())
            case _:
                return Boolean(False)

    def __rich_console__(
        self, console: Console, options: ConsoleOptions
    ) -> RenderResult:
        yield Segment(str(self._raw).lower(), style=STYLE_KEYWORDS)

    def rich_highlights(self) -> Iterable[Span]:
        if self.span is not None:
            yield self.span.styled(STYLE_KEYWORDS)


@dataclass(eq=True, frozen=True)
class Array(Value, MutableSequence[Value]):
    _raw: list[Value]

    def __equals__(self, other: Value) -> Boolean:
        match other:
            case Array() as other:
                return Boolean(self._raw == other._raw)
            case _:
                return Boolean(False)

    def insert(self, index: int, value: Value) -> None:
        self._raw.insert(index, value)

    @overload
    def __getitem__(self, index: int) -> Value: ...

    @overload
    def __getitem__(self, index: slice) -> MutableSequence[Value]: ...

    def __getitem__(self, index):
        return self._raw[index]

    @overload
    def __setitem__(self, index: int, value: Value) -> None: ...

    @overload
    def __setitem__(self, index: slice, value: Iterable[Value]) -> None: ...

    def __setitem__(self, index, value):
        self._raw[index] = value

    @overload
    def __delitem__(self, index: int) -> None: ...

    @overload
    def __delitem__(self, index: slice) -> None: ...

    def __delitem__(self, index):
        del self._raw[index]

    def __len__(self) -> int:
        return len(self._raw)

    def raw(self) -> list[object]:
        return [item.raw() for item in self._raw]

    def resolve(self) -> Value:
        unknown = []
        for item in self._raw:
            resolved_item = item.resolve()
            if isinstance(resolved_item, Unknown):
                unknown.append(resolved_item)
        if unknown:
            return Unknown.indirect(*unknown)
        return self

    def __rich_console__(
        self, console: Console, options: ConsoleOptions
    ) -> RenderResult:
        yield Segment("[")
        for i, item in enumerate(self._raw):
            yield item
            if i < len(self._raw) - 1:
                yield Segment(", ")

        yield Segment("]")


@dataclass(eq=True, frozen=True)
class Object(Value, MutableMapping[String, Value]):
    _raw: dict[String, Value]

    def __equals__(self, other: Value) -> Boolean:
        match other:
            case Object() as other:
                return Boolean(self._raw == other._raw)
            case _:
                return Boolean(False)

    def __setitem__(self, key: String, value: Value) -> None:
        self._raw[key] = value

    def __getitem__(self, key: String) -> Value:
        return self._raw[key]

    def __delitem__(self, key: String) -> None:
        del self._raw[key]

    def __len__(self) -> int:
        return len(self._raw)

    def __iter__(self) -> Iterator[String]:
        return iter(self._raw)

    def raw(self) -> dict[object, object]:
        return {key.raw(): value.raw() for key, value in self._raw.items()}

    def resolve(self) -> Value:
        unknown = []
        for key, value in self._raw.items():
            resolved_value = value.resolve()
            if isinstance(resolved_value, Unknown):
                unknown.append(resolved_value)
        if unknown:
            return Unknown.indirect(*unknown)
        return self

    def __rich_console__(
        self, console: Console, options: ConsoleOptions
    ) -> RenderResult:
        yield Segment("{")
        for i, (key, value) in enumerate(self._raw.items()):
            yield Segment(key.raw(), style=STYLE_PROPERTY_NAME)
            yield Segment(" = ")
            yield value
            if i < len(self._raw) - 1:
                yield Segment(", ")

        yield Segment("}")


@dataclass(eq=True, frozen=True)
class VariableReference:
    key: tuple[str | None, ...]
    span: SourceSpan


@dataclass(eq=True, frozen=True)
class Unknown(Value):
    # Direct refs are the missing variable/path we tried to read. Examples:
    # - `var.foo` => direct: var.foo
    # - `obj.bar` => direct: obj.bar
    # - `obj["baz"]` => direct: obj.baz
    # If we access attributes on an unknown, the path extends:
    # `var.foo.bar` => direct: var.foo.bar
    # The base unknown remains indirect (e.g., `var.foo` in the example above).
    # Indirect refs are unknowns that only block evaluation of another value:
    # - `var.foo + 1` => indirect: var.foo
    # - `cond ? x : y` => indirect: cond (plus x/y if unknown)
    direct_references: set[VariableReference] = field(default_factory=set)
    indirect_references: set[VariableReference] = field(default_factory=set)

    def raise_on_unknown(self) -> Never:
        raise DiagnosticError(
            code="pyhcl2::evaluator::unknown_variable",
            message=Inline("Failed to evaluate expression due to unknown variables"),
            labels=[
                LabeledSpan(
                    ref.span,
                    f"{ref.key[-1]} could not be resolved ({'.'.join([k if k else '?' for k in ref.key])})",
                )
                for ref in self.references
            ],
        )

    def raw(self) -> Never:
        self.raise_on_unknown()

    @property
    def references(self) -> set[VariableReference]:
        return self.direct_references | self.indirect_references

    def __rich_console__(
        self, console: Console, options: ConsoleOptions
    ) -> RenderResult:
        yield Segment("<")
        yield Segment(
            "Unknown due to missing variables, direct: ", style=STYLE_KEYWORDS
        )
        for i, ref in enumerate(self.direct_references):
            for j, key in enumerate(ref.key):
                yield Segment(key if key else "?")
                if j < len(ref.key) - 1:
                    yield Segment(".")
            if i < len(self.direct_references) - 1:
                yield Segment(", ")
        yield Segment(", indirect: ", style=STYLE_KEYWORDS)
        for i, ref in enumerate(self.indirect_references):
            for j, key in enumerate(ref.key):
                yield Segment(key if key else "?")
                if j < len(ref.key) - 1:
                    yield Segment(".")
            if i < len(self.indirect_references) - 1:
                yield Segment(", ")
        yield Segment(">")

    def indirect(*values: Value) -> Unknown:
        resolved_values = [value.resolve() for value in values]

        return Unknown(
            set(),
            set(
                [
                    ref
                    for value in resolved_values
                    if isinstance(value, Unknown)
                    for ref in value.references
                ]
            ),
        )

    def direct(self, span: SourceSpan, key: str) -> Unknown:
        if self.direct_references:
            direct_refs = set(
                [
                    VariableReference(
                        (
                            *ref.key,
                            key,
                        ),
                        span,
                    )
                    for ref in self.direct_references
                ]
            )
        else:
            direct_refs = {
                VariableReference(
                    (
                        None,
                        key,
                    ),
                    span,
                )
            }

        return Unknown(direct_refs, self.references)

    @staticmethod
    def ident(identifier: pyhcl2.nodes.Identifier) -> Unknown:
        return Unknown(
            {VariableReference((identifier.name,), identifier.span)},
            set(),
            span=identifier.span,
        )



from __future__ import annotations

import re
from collections.abc import Callable, Iterable, Mapping, MutableMapping
from dataclasses import dataclass, field
from typing import Self

from pyagnostics.exceptions import DiagnosticError
from pyagnostics.spans import LabeledSpan, SourceSpan
from rich.console import Group
from rich.text import Text

from pyhcl2.nodes import (
    ArrayExpression,
    Attribute,
    AttrSplat,
    BinaryExpression,
    Block,
    Conditional,
    ForObjectExpression,
    ForTupleExpression,
    FunctionCall,
    GetAttr,
    GetAttrKey,
    GetIndex,
    GetIndexKey,
    Identifier,
    IndexSplat,
    Literal,
    Node,
    ObjectExpression,
    Parenthesis,
    UnaryExpression,
)
from pyhcl2.rich_utils import Inline
from pyhcl2.values import Array, Boolean, Integer, Null, Object, String, Unknown, Value

camel_to_snake_pattern = re.compile(r"(?<!^)(?=[A-Z])")


@dataclass
class EvaluationScope:
    parent: Self | None = None
    variables: MutableMapping[str, Value] = field(default_factory=dict)

    def __post_init__(self) -> None:
        for key, value in self.variables.items():
            if not isinstance(value, Value):
                raise TypeError(f"Variable {key} is not a Value")

    def __getitem__(self, item: str) -> Value:
        try:
            return self.variables[item]
        except KeyError:
            pass
        if self.parent:
            return self.parent[item]
        raise KeyError(f"Variable {item} not set")

    def __setitem__(self, key: str, value: Value) -> None:
        self.variables[key] = value

    def __contains__(self, item: Value) -> bool:
        if item in self.variables:
            return True
        if self.parent:
            return item in self.parent
        return False

    def child(self) -> EvaluationScope:
        return EvaluationScope(parent=self)


@dataclass
class Evaluator:
    intrinsic_functions: Mapping[str, Callable[..., Value]] = field(
        default_factory=dict
    )

    def eval(self, expr: Node, scope: EvaluationScope | None = None) -> Value:  # noqa: PLR0912
        scope = scope or EvaluationScope()
        match expr:
            case Block() as expr:
                result = self._eval_block(expr, scope)
            case Literal() as expr:
                result = self._eval_literal(expr, scope)
            case ArrayExpression() as expr:
                result = self._eval_array_expression(expr, scope)
            case ObjectExpression() as expr:
                result = self._eval_object_expression(expr, scope)
            case Identifier() as expr:
                result = self._eval_identifier(expr, scope)
            case Parenthesis() as expr:
                result = self._eval_parenthesis(expr, scope)
            case BinaryExpression() as expr:
                result = self._eval_binary_expression(expr, scope)
            case UnaryExpression() as expr:
                result = self._eval_unary_expression(expr, scope)
            case Attribute() as expr:
                result = self._eval_attribute(expr, scope)
            case GetAttr() as expr:
                result = self._eval_get_attr(expr, scope)
            case GetIndex() as expr:
                result = self._eval_get_index(expr, scope)
            case FunctionCall() as expr:
                result = self._eval_function_call(expr, scope)
            case Conditional() as expr:
                result = self._eval_conditional(expr, scope)
            case ForTupleExpression() as expr:
                result = self._eval_for_tuple_expression(expr, scope)
            case ForObjectExpression() as expr:
                result = self._eval_for_object_expression(expr, scope)
            case AttrSplat() as expr:
                result = self._eval_attr_splat(expr, scope)
            case IndexSplat() as expr:
                result = self._eval_index_splat(expr, scope)
            case _:
                raise DiagnosticError(
                    code="pyhcl2::evaluator::unsupported_node",
                    message=f"Unsupported node type {expr.__class__.__name__}",
                    labels=[
                        LabeledSpan(expr.span, "unsupported expression"),
                    ],
                )

        # rich.print(Inline(Text("eval", style=STYLE_FUNCTION), "(", expr, "):  ", result, NewLine()))
        if result.span is None:
            result = result.with_span(expr.span)
        return result

    def _eval_block(self, block: Block, scope: EvaluationScope) -> Value:
        result: dict[String, Value] = {}

        for stmt in block.body:
            match stmt:
                case Attribute() as attr:
                    key = attr.key.as_string()
                    value = self.eval(stmt, scope.child())

                    if key in result:
                        raise DiagnosticError(
                            code="pyhcl2::evaluator::block::duplicate_key",
                            message="Duplicate key in block",
                            labels=[LabeledSpan(attr.key.span, "duplicate key")],
                        )

                    result[key] = value

                case Block() as block:
                    keys = block.keys
                    value = self.eval(block, scope.child())

                    mapping: MutableMapping[String, Value] = result
                    for key in keys[:-1]:
                        match mapping.get(key, None):
                            case None:
                                mapping = mapping[key] = Object({})
                            case Object() as obj:
                                mapping = obj
                            case _:
                                raise DiagnosticError(
                                    code="pyhcl2::evaluator::block::key_conflict",
                                    message="Key conflict in block",
                                    labels=[LabeledSpan(block.span, "key conflict")],
                                )

                    match mapping.get(keys[-1], None):
                        case None:
                            mapping[keys[-1]] = Array([value])
                        case Array() as array:
                            array.append(value)
                        case _:
                            raise DiagnosticError(
                                code="pyhcl2::evaluator::block::key_conflict",
                                message="Key conflict in block",
                                labels=[LabeledSpan(block.span, "key conflict")],
                            )

        return Object(result)

    @staticmethod
    def _eval_literal(expr: Literal, _scope: EvaluationScope) -> Value:
        return expr.value

    def _eval_array_expression(
        self, expr: ArrayExpression, scope: EvaluationScope
    ) -> Value:
        return Array([self.eval(item, scope) for item in expr.values])

    def _eval_object_expression(
        self, obj: ObjectExpression, scope: EvaluationScope
    ) -> Value:
        result: dict[String, Value] = {}

        unknown_keys = []

        for key_expr, value_expr in obj.fields.items():
            resolved_key: String
            match key_expr:
                case Identifier(name):
                    resolved_key = String(name)
                case Literal(String() as string):
                    resolved_key = string
                case Parenthesis(expr):
                    key = self.eval(expr, scope)

                    match key:
                        case Unknown() as key:
                            unknown_keys.append(key)
                            continue
                        case String() as key:
                            resolved_key = key
                        case _:
                            raise DiagnosticError(
                                code="pyhcl2::evaluator::object::unsupported_key",
                                message="Unsupported key type in object",
                                labels=[LabeledSpan(expr.span, key.type_name)],
                            )
                case _:
                    raise DiagnosticError(
                        code="pyhcl2::evaluator::object::unsupported_key",
                        message="Unsupported key type in object",
                        labels=[LabeledSpan(key_expr.span, "unsupported key")],
                        notes=[
                            Inline(
                                "[blue]help:[/blue] ",
                                "Did you mean `",
                                Parenthesis(key_expr),
                                " = ",
                                value_expr,
                                "`?",
                            )
                        ],
                    )
            value = self.eval(value_expr, scope)
            if isinstance(value, Unknown):
                value = value.indirect()
            result[resolved_key] = value

        if unknown_keys:
            return Unknown.indirect(*unknown_keys)

        return Object(result)

    @staticmethod
    def _eval_identifier(identifier: Identifier, scope: EvaluationScope) -> Value:
        try:
            return scope[identifier.name]
        except KeyError:
            return Unknown.ident(identifier)

    def _eval_parenthesis(self, paren: Parenthesis, scope: EvaluationScope) -> Value:
        return self.eval(paren.expr, scope)

    def _eval_binary_expression(
        self, expr: BinaryExpression, scope: EvaluationScope
    ) -> Value:
        operation = {
            "+": "__add__",
            "-": "__sub__",
            "*": "__mul__",
            "/": "__truediv__",
            "%": "__mod__",
            "==": "__equals__",
            "!=": "__not_equals__",
            "<": "__lt__",
            ">": "__gt__",
            "<=": "__le__",
            ">=": "__ge__",
            "&&": "__and__",
            "||": "__or__",
        }[expr.op.type]

        left = self.eval(expr.left, scope)
        try:
            if isinstance(left, Unknown):
                return Unknown.indirect(left, self.eval(expr.right, scope))

            operator = getattr(left, operation)
            right = self.eval(expr.right, scope)

            if isinstance(right, Unknown):
                return right

            result = operator(right)

            if result is NotImplemented:
                raise NotImplementedError()  # noqa: TRY301
            else:
                return result

        except (AttributeError, NotImplementedError):
            right = self.eval(expr.right, scope)
            raise DiagnosticError(
                code="pyhcl2::evaluator::binary_expression::unsupported_operator",
                message=f"Binary operator `{expr.op.type}` not implemented for operands of types {left.type_name} and {right.type_name}",
                labels=[
                    LabeledSpan(expr.left.span, left.type_name),
                    LabeledSpan(expr.op.span, "unsupported operator"),
                    LabeledSpan(expr.right.span, right.type_name),
                ],
            ) from None
        except ArithmeticError as e:
            right = self.eval(expr.right, scope)
            raise DiagnosticError(
                code="pyhcl2::evaluator::binary_expression::arithmetic_error",
                message=f"An {e} error occurred",
                labels=[
                    LabeledSpan(
                        expr.left.span, Group(Text("left operand:", end=" "), left)
                    ),
                    LabeledSpan(expr.op.span, str(e)),
                    LabeledSpan(
                        expr.right.span, Group(Text("right operand:", end=" "), right)
                    ),
                ],
            ).with_context("while evaluating binary expression")

    def _eval_unary_expression(
        self, expr: UnaryExpression, scope: EvaluationScope
    ) -> Value:
        value = self.eval(expr.expr, scope)

        try:
            operation = {
                "-": "__neg__",
                "!": "__not__",
            }[expr.op.type]
            result = getattr(value, operation)()

            if result is NotImplemented:
                raise NotImplementedError()  # noqa: TRY301
            else:
                return result

        except (AttributeError, NotImplementedError):
            raise DiagnosticError(
                code="pyhcl2::evaluator::unsupported_unary_operator",
                message=f"Unary operator `{expr.op.type}` not implemented for operand of type {value.type_name}",
                labels=[
                    LabeledSpan(expr.op.span, "unsupported operator"),
                    LabeledSpan(expr.expr.span, value.type_name),
                ],
            )

    def _eval_attribute(self, attr: Attribute, scope: EvaluationScope) -> Value:
        value = self.eval(attr.value, scope)
        scope[attr.key.name] = value
        return value

    def _eval_get_attr(self, expr: GetAttr, scope: EvaluationScope) -> Value:
        on = self.eval(expr.on, scope)
        return self._evaluate_get_attr(on, expr.on.span, expr.key, scope)

    def _eval_get_index(self, expr: GetIndex, scope: EvaluationScope) -> Value:
        on = self.eval(expr.on, scope)
        return self._evaluate_get_index(on, expr.on.span, expr.key, scope)

    @staticmethod
    def _evaluate_get_attr(
        on: Value, on_span: SourceSpan, key: GetAttrKey, _scope: EvaluationScope
    ) -> Value:
        key_value = key.ident.name

        if isinstance(on, Unknown):
            return on.direct(key.ident.span, key.ident.name)

        if not isinstance(on, Object):
            raise DiagnosticError(
                code="pyhcl2::evaluator::get_attr::unsupported_type",
                message="Cannot get attribute from non-object type",
                labels=[
                    LabeledSpan(on_span, on.type_name),
                    LabeledSpan(key.ident.span, "unsupported attribute"),
                ],
            )

        try:
            return on[String(key_value)]
        except KeyError:
            return Unknown().direct(key.ident.span, key_value)

    def _evaluate_get_index(
        self, on: Value, on_span: SourceSpan, key: GetIndexKey, scope: EvaluationScope
    ) -> Value:
        key_value = self.eval(key.expr, scope)

        try:
            match on, key_value:
                case Unknown() as on, Unknown() as key_value:
                    return Unknown.indirect(on, key_value)
                case Unknown() as on, String(raw_str):
                    return on.direct(key.expr.span, raw_str)
                case Unknown() as on, Integer():
                    return Unknown.indirect(on, key_value)
                case Object() as obj, String() as key_value:
                    return obj[key_value]
                case Array() as on, Integer(int_raw):
                    return on[int_raw]
                case Object(), Unknown() as key_value:
                    return key_value
                case _:
                    raise DiagnosticError(
                        code="pyhcl2::evaluator::get_index::unsupported_type",
                        message=f"Cannot index into {on.type_name} with {key_value.type_name} key",
                        labels=[
                            LabeledSpan(on_span, on.type_name),
                            LabeledSpan(key.expr.span, key_value.type_name),
                        ],
                    )
        except KeyError:
            raise DiagnosticError(
                code="pyhcl2::evaluator::get_index::missing_key",
                message="Key not found in object",
                labels=[
                    LabeledSpan(on_span, on.type_name),
                    LabeledSpan(key.expr.span, "key"),
                ],
            )
        except IndexError:
            raise DiagnosticError(
                code="pyhcl2::evaluator::get_index::index_out_of_bounds",
                message="Index out of bounds",
                labels=[
                    LabeledSpan(on_span, on.type_name),
                    LabeledSpan(key.expr.span, "index"),
                ],
            )

    def _eval_function_call(self, call: FunctionCall, scope: EvaluationScope) -> Value:
        if call.var_args:
            # TODO
            raise DiagnosticError(
                code="pyhcl2::evaluator::function_call::unsupported_var_args",
                message="Function calls with var args are not supported",
                labels=[LabeledSpan(call.span, "call with var args")],
            )

        if call.ident.name in self.intrinsic_functions:
            args = [self.eval(arg, scope) for arg in call.args]
            if any(isinstance(arg.resolve(), Unknown) for arg in args):
                return Unknown.indirect(*args)

            try:
                result = self.intrinsic_functions[call.ident.name](*args)
                if result is None:
                    raise DiagnosticError(
                        code="pyhcl2::evaluator::function_call::invalid_value",
                        message="Intrinsic function returned invalid value",
                        labels=[LabeledSpan(call.ident.span, "intrinsic function")],
                    )
                else:
                    return result
            except TypeError as e:
                raise DiagnosticError(
                    code="pyhcl2::evaluator::function_call::invalid_args",
                    message="Invalid arguments passed to function",
                    labels=[LabeledSpan(call.args_span, "invalid arguments")],
                ) from e

        raise DiagnosticError(
            code="pyhcl2::evaluator::function_call::unsupported_function",
            message=f"Intrinsic function `{call.ident.name}` does not exist",
            labels=[LabeledSpan(call.ident.span, "unsupported function")],
        )

    def _eval_conditional(self, expr: Conditional, scope: EvaluationScope) -> Value:
        condition = self.eval(expr.cond, scope)

        match condition:
            case Unknown() as condition:
                return Unknown.indirect(
                    condition,
                    self.eval(expr.then_expr, scope),
                    self.eval(expr.else_expr, scope),
                )
            case Boolean(True):
                return self.eval(expr.then_expr, scope)
            case Boolean(False):
                return self.eval(expr.else_expr, scope)
            case _:
                raise DiagnosticError(
                    code="pyhcl2::evaluator::conditional::unsupported_condition",
                    message=f"Unsupported condition type {condition.type_name}",
                    labels=[LabeledSpan(expr.cond.span, condition.type_name)],
                )

    def _eval_for_tuple_expression(
        self, expr: ForTupleExpression, scope: EvaluationScope
    ) -> Value:
        collection = self.eval(expr.collection, scope)
        results: list[Value] = []

        iterator: Iterable[tuple[Value, Value]]

        match collection:
            case Object(obj):
                iterator = obj.items()
            case Array(array):
                iterator = [(Integer(k), v) for k, v in enumerate(array)]
            case Unknown() as collection:
                unknown = collection.indirect()
                iterator = [(unknown, unknown)]
            case _:
                raise DiagnosticError(
                    code="pyhcl2::evaluator::for_tuple_expression::unsupported_collection",
                    message=f"Unsupported collection type {collection.type_name}",
                    labels=[LabeledSpan(expr.collection.span, collection.type_name)],
                )

        for k, v in iterator:
            child_scope = scope.child()
            child_scope[expr.value_ident.name] = v
            if expr.key_ident:
                child_scope[expr.key_ident.name] = k

            condition = (
                self.eval(expr.condition, child_scope)
                if expr.condition is not None
                else Boolean(True)
            )

            match condition:
                case Unknown() as condition:
                    results.append(
                        Unknown.indirect(condition, self.eval(expr.value, child_scope))
                    )
                case Boolean(True):
                    result = self.eval(expr.value, child_scope)
                    match result:
                        case Unknown() as result:
                            results.append(result.indirect())
                        case _:
                            results.append(result)
                case Boolean(False):
                    pass
                case _:
                    assert expr.condition is not None
                    raise DiagnosticError(
                        code="pyhcl2::evaluator::for_tuple_expression::unsupported_condition",
                        message=f"Unsupported condition type {condition.type_name}",
                        labels=[LabeledSpan(expr.condition.span, condition.type_name)],
                    )

        return Array(results)

    def _eval_for_object_expression(  # noqa: PLR0912
        self, expr: ForObjectExpression, scope: EvaluationScope
    ) -> Value:
        if expr.grouping_mode:
            raise DiagnosticError(
                code="pyhcl2::evaluator::for_object_expression::unsupported_grouping_mode",
                message="Grouping mode is not supported",
                labels=[LabeledSpan(expr.span, "grouping mode")],
            )

        collection = self.eval(expr.collection, scope)
        results: dict[String, Value] = {}
        unknown_blockers = []

        iterator: Iterable[tuple[Value, Value]]

        match collection:
            case Object(obj):
                iterator = obj.items()
            case Array(array):
                iterator = [(Integer(k), v) for k, v in enumerate(array)]
            case Unknown() as collection:
                unknown = collection.indirect()
                iterator = [(unknown, unknown)]
            case _:
                raise DiagnosticError(
                    code="pyhcl2::evaluator::for_object_expression::unsupported_collection",
                    message=f"Unsupported collection type {collection.type_name}",
                    labels=[LabeledSpan(expr.collection.span, collection.type_name)],
                )

        for k, v in iterator:
            child_scope = scope.child()
            child_scope[expr.value_ident.name] = v
            if expr.key_ident:
                child_scope[expr.key_ident.name] = k

            condition = (
                self.eval(expr.condition, child_scope)
                if expr.condition is not None
                else Boolean(True)
            )

            match condition:
                case Unknown() as condition:
                    unknown_blockers.append(
                        Unknown.indirect(
                            condition,
                            self.eval(expr.key, child_scope),
                            self.eval(expr.value, child_scope),
                        )
                    )
                case Boolean(True):
                    key = self.eval(expr.key, child_scope)
                    match key:
                        case Unknown() as key:
                            unknown_blockers.append(
                                Unknown.indirect(
                                    key, self.eval(expr.value, child_scope)
                                )
                            )
                        case String() as key:
                            results[key] = self.eval(expr.value, child_scope)
                        case _:
                            raise DiagnosticError(
                                code="pyhcl2::evaluator::for_object_expression::unsupported_key",
                                message=f"Unsupported key type {key.type_name}",
                                labels=[LabeledSpan(expr.key.span, key.type_name)],
                            )
                case Boolean(False):
                    pass
                case _:
                    assert expr.condition is not None
                    raise DiagnosticError(
                        code="pyhcl2::evaluator::for_object_expression::unsupported_condition",
                        message=f"Unsupported condition type {condition.type_name}",
                        labels=[LabeledSpan(expr.condition.span, condition.type_name)],
                    )

        if unknown_blockers:
            return Unknown.indirect(*unknown_blockers)

        return Object(results)

    def _eval_attr_splat(self, expr: AttrSplat, scope: EvaluationScope) -> Value:
        on = self.eval(expr.on, scope)

        match on:
            case Null():
                return Array([])
            case Array(array):
                iterable = array
            case _:
                iterable = [on]

        values = []

        for i, v in enumerate(iterable):
            try:
                span = expr.on.span
                value = v
                for key in expr.keys:
                    value = self._evaluate_get_attr(value, span, key, scope)
                    span = SourceSpan(span.start, key.span.end)
                values.append(value)
            except DiagnosticError as e:
                e.notes.append(
                    Inline(
                        "[blue]help:[/blue] The resulting expression was ",
                        v,
                        *expr.keys,
                    )
                )
                raise e.with_context(f"while evaluating element {i}").with_context(
                    "while evaluating attribute splat expression"
                )
        if isinstance(on, Unknown):
            return Unknown.indirect(*values)

        return Array(values)

    def _eval_index_splat(self, expr: IndexSplat, scope: EvaluationScope) -> Value:
        on = self.eval(expr.on, scope)

        match on:
            case Null():
                return Array([])
            case Array(array):
                iterable = array
            case _:
                iterable = [on]

        values = []

        for i, v in enumerate(iterable):
            try:
                span = expr.on.span
                value = v
                for key in expr.keys:
                    match key:
                        case GetAttrKey():
                            value = self._evaluate_get_attr(value, span, key, scope)
                        case GetIndexKey():
                            value = self._evaluate_get_index(value, span, key, scope)
                    span = SourceSpan(span.start, key.span.end)
                values.append(value)
            except DiagnosticError as e:
                e.notes.append(
                    Inline(
                        "[blue]help:[/blue] The resulting expression was ",
                        v,
                        *expr.keys,
                    )
                )
                raise e.with_context(f"while evaluating element {i}").with_context(
                    "while evaluating index splat expression"
                )

        if isinstance(on, Unknown):
            return Unknown.indirect(*values)

        return Array(values)



from __future__ import annotations

import sys
from collections.abc import Callable, Iterator, Mapping
from pathlib import Path
from typing import (
    Self,
    cast,
)

import rich
from rich.console import NewLine

from pyhcl2.eval import EvaluationScope, Evaluator
from pyhcl2.nodes import Block, Node
from pyhcl2.parse import parse_file
from pyhcl2.rich_utils import Inline
from pyhcl2.values import Unknown, Value


class IntrinsicFunctionTracker(Mapping):
    def __getattr__(self, item: str) -> Callable[..., Value]:
        return lambda *args: Unknown.indirect(*args)

    def __getitem__(self, item: str) -> Callable[..., Value]:
        return lambda *args: Unknown.indirect(*args)

    def __contains__(self, item: object) -> bool:
        return True

    def __iter__(self) -> Iterator[Self]:
        return iter(())

    def __len__(self) -> int:
        return 0


def resolve_variable_references(node: Node) -> set[tuple[str, ...]]:
    # noinspection PyTypeChecker
    scope = EvaluationScope()
    result = Evaluator(intrinsic_functions=IntrinsicFunctionTracker()).eval(node, scope)

    match result.resolve():
        case Unknown() as unknown:
            return set(
                [
                    cast(tuple[str, ...], ref.key)
                    for ref in unknown.references
                    if not any(part is None for part in ref.key)
                ]
            )
        case _:
            return set()


if __name__ == "__main__":
    ast = parse_file(open(Path(sys.argv[1])))
    rich.print(ast)

    blocks = [stmt for stmt in ast.body if isinstance(stmt, Block)]

    for block_under_test in blocks:
        variable_references = resolve_variable_references(block_under_test)

        for dirty_child in variable_references:
            rich.print(
                Inline(
                    ".".join(block_under_test.key()),
                    " depends_on ",
                    ".".join(dirty_child),
                    NewLine(),
                )
            )



# ruff: noqa: PLC0415
from pathlib import Path

import rich
from pyagnostics.exceptions import DiagnosticError
from pyagnostics.source import InMemorySource, attach_diagnostic_source_code
from rich.console import NewLine

from pyhcl2.eval import EvaluationScope, Evaluator
from pyhcl2.parse import parse_expr_or_stmt
from pyhcl2.rich_utils import HclHighlighter, Inline


def main() -> None:
    from prompt_toolkit import PromptSession  # type: ignore
    from prompt_toolkit.auto_suggest import AutoSuggestFromHistory  # type: ignore
    from prompt_toolkit.history import FileHistory  # type: ignore

    scope = EvaluationScope()
    evaluator = Evaluator(intrinsic_functions={"identity": lambda x: x})
    session: PromptSession = PromptSession(
        history=FileHistory(Path.home() / ".pyhcl2_history")
    )
    try:
        while True:
            text = session.prompt("> ", auto_suggest=AutoSuggestFromHistory())

            if text == "":
                continue

            if text == "exit":
                break

            src = InMemorySource(text)
            try:
                ast = parse_expr_or_stmt(text)
                with attach_diagnostic_source_code(
                    src, highlighter=HclHighlighter(ast)
                ):
                    result = evaluator.eval(ast, scope)
                    rich.print(Inline(result.resolve().raise_on_unknown(), NewLine()))
            except DiagnosticError as diagnostic:
                rich.get_console().print(
                    diagnostic.with_source_code(src), highlight=False
                )

    except KeyboardInterrupt:
        pass


if __name__ == "__main__":
    main()



from __future__ import annotations

from dataclasses import dataclass

from pyagnostics.protocols import SourceCodeHighlighter, SpanContents
from rich.color import Color
from rich.console import Group, RenderableType
from rich.style import Style
from rich.text import Span, Text

from pyhcl2 import nodes

STYLE_PROPERTY_NAME = Style(color=Color.from_rgb(199, 125, 187))
STYLE_KEYWORDS = Style(color=Color.from_rgb(207, 142, 109))
STYLE_NUMBER = Style(color=Color.from_rgb(42, 172, 184))
STYLE_STRING = Style(color=Color.from_rgb(106, 171, 115))
STYLE_FUNCTION = Style(color=Color.from_rgb(136, 136, 198))


@dataclass
class HclHighlighter(SourceCodeHighlighter):
    ast: nodes.Node

    def highlight(self, span_contents: SpanContents) -> SpanContents:
        spans: list[Span] = [
            Span(
                span.start - span_contents.span.start,
                span.end - span_contents.span.start,
                span.style,
            )
            for span in self.ast.rich_highlights()
            if span.start < span_contents.span.end
            and span.end > span_contents.span.start
        ]
        span_contents.text.spans.extend(spans)
        return span_contents


class Inline:
    def __init__(self, *renderables: RenderableType) -> None:
        self.renderables: list[RenderableType] = []
        for renderable in renderables:
            match renderable:
                case str() as text:
                    self.renderables.append(Text(text, end=""))
                case Text() as text:
                    text = text.copy()
                    text.end = ""
                    self.renderables.append(text)
                case _:
                    self.renderables.append(renderable)

    def __rich__(self) -> Group:
        return Group(*self.renderables)



from __future__ import annotations

import dataclasses
import typing as t
from collections.abc import Iterable
from dataclasses import dataclass, field
from functools import cached_property

from pyagnostics.spans import SourceSpan
from rich.console import (
    Console,
    ConsoleOptions,
    ConsoleRenderable,
    RenderResult,
)
from rich.padding import Padding
from rich.segment import Segment
from rich.text import Span

from pyhcl2.rich_utils import (
    STYLE_FUNCTION,
    STYLE_KEYWORDS,
    STYLE_PROPERTY_NAME,
)
from pyhcl2.values import String, Value


@dataclass(frozen=True, eq=True, kw_only=True)
class Node(ConsoleRenderable):
    """Base class for HCL2 AST nodes."""

    span: SourceSpan = field(default=SourceSpan(-1, -1), compare=False, hash=False)

    def rich_highlights(self) -> Iterable[Span]:
        return []


class Expression(Node):
    """Base class for nodes that represent expressions in HCL2."""


@dataclass(frozen=True, eq=True)
class Literal(Expression):
    value: Value

    def __rich_console__(
        self, console: Console, options: ConsoleOptions
    ) -> RenderResult:
        yield self.value

    def rich_highlights(self) -> Iterable[Span]:
        return self.value.rich_highlights()


@dataclass(frozen=True, eq=True)
class ArrayExpression(Expression):
    values: list[Expression]

    def __rich_console__(
        self, console: Console, options: ConsoleOptions
    ) -> RenderResult:
        yield Segment("[")
        for i, value in enumerate(self.values):
            yield value
            if i < len(self.values) - 1:
                yield Segment(", ")
        yield Segment("]")

    def rich_highlights(self) -> Iterable[Span]:
        for value in self.values:
            yield from value.rich_highlights()


@dataclass(frozen=True, eq=True)
class ObjectExpression(Expression):
    fields: dict[Expression, Expression]

    def __rich_console__(
        self, console: Console, options: ConsoleOptions
    ) -> RenderResult:
        yield Segment("{")
        for i, (key, value) in enumerate(self.fields.items()):
            if isinstance(key, Identifier):
                yield Segment(key.name, style=STYLE_PROPERTY_NAME)
            else:
                yield key
            yield Segment(" = ")
            yield value
            if i < len(self.fields) - 1:
                yield Segment(", ")
        yield Segment("}")

    def rich_highlights(self) -> Iterable[Span]:
        for key, value in self.fields.items():
            if isinstance(key, Identifier):
                yield key.span.styled(STYLE_PROPERTY_NAME)
            else:
                yield from key.rich_highlights()
            yield from value.rich_highlights()


@dataclass(frozen=True, eq=True)
class Identifier(Expression):
    name: str

    def as_string(self) -> String:
        return String(
            self.name,
            span=self.span,
        )

    def __rich_console__(
        self, console: Console, options: ConsoleOptions
    ) -> RenderResult:
        yield Segment(self.name)

    def rich_highlights(self) -> Iterable[Span]:
        return []


@dataclass(frozen=True, eq=True)
class FunctionCall(Expression):
    ident: Identifier
    args: list[Expression]
    var_args: bool = False

    @property
    def args_span(self) -> SourceSpan:
        return SourceSpan(self.ident.span.end, self.span.end)

    def __post_init__(self) -> None:
        assert all(isinstance(arg, Expression) for arg in self.args), self.args

    def __rich_console__(
        self, console: Console, options: ConsoleOptions
    ) -> RenderResult:
        yield Segment(self.ident.name, style=STYLE_FUNCTION)
        yield Segment("(")
        for i, arg in enumerate(self.args):
            yield arg
            if i < len(self.args) - 1:
                yield Segment(", ")
        if self.var_args:
            yield Segment("...")
        yield Segment(")")

    def rich_highlights(self) -> Iterable[Span]:
        yield self.ident.span.styled(STYLE_FUNCTION)
        for arg in self.args:
            yield from arg.rich_highlights()


@dataclass(frozen=True, eq=True)
class GetAttrKey(Node):
    ident: Identifier

    def __rich_console__(
        self, console: Console, options: ConsoleOptions
    ) -> RenderResult:
        yield Segment(".")
        yield self.ident


@dataclass(frozen=True, eq=True)
class GetIndexKey(Node):
    expr: Expression

    def __rich_console__(
        self, console: Console, options: ConsoleOptions
    ) -> RenderResult:
        yield Segment("[")
        yield self.expr
        yield Segment("]")

    def rich_highlights(self) -> Iterable[Span]:
        yield from self.expr.rich_highlights()


@dataclass(frozen=True, eq=True)
class GetAttr(Expression):
    on: Expression
    key: GetAttrKey

    def __rich_console__(
        self, console: Console, options: ConsoleOptions
    ) -> RenderResult:
        yield self.on
        yield self.key

    def rich_highlights(self) -> Iterable[Span]:
        yield from self.on.rich_highlights()
        yield from self.key.ident.rich_highlights()


@dataclass(frozen=True, eq=True)
class GetIndex(Expression):
    on: Expression
    key: GetIndexKey

    def __rich_console__(
        self, console: Console, options: ConsoleOptions
    ) -> RenderResult:
        yield self.on
        yield self.key

    def rich_highlights(self) -> Iterable[Span]:
        yield from self.on.rich_highlights()
        yield from self.key.expr.rich_highlights()


@dataclass(frozen=True, eq=True)
class AttrSplat(Expression):
    on: Expression
    keys: list[GetAttrKey] = dataclasses.field(default_factory=list)

    def __rich_console__(
        self, console: Console, options: ConsoleOptions
    ) -> RenderResult:
        yield self.on
        yield Segment(".*")
        yield from self.keys

    def rich_highlights(self) -> Iterable[Span]:
        yield from self.on.rich_highlights()
        for key in self.keys:
            yield from key.ident.rich_highlights()


@dataclass(frozen=True, eq=True)
class IndexSplat(Expression):
    on: Expression
    keys: list[GetAttrKey | GetIndexKey] = dataclasses.field(default_factory=list)

    def __rich_console__(
        self, console: Console, options: ConsoleOptions
    ) -> RenderResult:
        yield self.on
        yield Segment("[*]")
        yield from self.keys

    def rich_highlights(self) -> Iterable[Span]:
        yield from self.on.rich_highlights()
        for key in self.keys:
            yield from key.rich_highlights()


@dataclass(frozen=True, eq=True)
class UnaryOperator(Node):
    type: t.Literal["-", "!"]

    def __rich_console__(
        self, console: Console, options: ConsoleOptions
    ) -> RenderResult:
        yield Segment(self.type)

    def rich_highlights(self) -> Iterable[Span]:
        return []


@dataclass(frozen=True, eq=True)
class UnaryExpression(Expression):
    op: UnaryOperator
    expr: Expression

    def __rich_console__(
        self, console: Console, options: ConsoleOptions
    ) -> RenderResult:
        yield self.op
        yield self.expr

    def rich_highlights(self) -> Iterable[Span]:
        yield from self.op.rich_highlights()
        yield from self.expr.rich_highlights()


@dataclass(frozen=True, eq=True)
class BinaryOperator(Node):
    type: t.Literal[
        "==", "!=", "<", ">", "<=", ">=", "-", "*", "/", "%", "&&", "||", "+"
    ]

    def __rich_console__(
        self, console: Console, options: ConsoleOptions
    ) -> RenderResult:
        yield Segment(self.type)

    def rich_highlights(self) -> Iterable[Span]:
        return []


@dataclass(frozen=True, eq=True)
class BinaryExpression(Expression):
    op: BinaryOperator
    left: Expression
    right: Expression

    def __rich_console__(
        self, console: Console, options: ConsoleOptions
    ) -> RenderResult:
        yield self.left
        yield Segment(" ")
        yield self.op
        yield Segment(" ")
        yield self.right

    def rich_highlights(self) -> Iterable[Span]:
        yield from self.left.rich_highlights()
        yield from self.op.rich_highlights()
        yield from self.right.rich_highlights()


@dataclass(frozen=True, eq=True)
class Conditional(Expression):
    cond: Expression
    then_expr: Expression
    else_expr: Expression

    def __rich_console__(
        self, console: Console, options: ConsoleOptions
    ) -> RenderResult:
        yield self.cond
        yield Segment(" ? ")
        yield self.then_expr
        yield Segment(" : ")
        yield self.else_expr

    def rich_highlights(self) -> Iterable[Span]:
        yield from self.cond.rich_highlights()
        yield from self.then_expr.rich_highlights()
        yield from self.else_expr.rich_highlights()


@dataclass(frozen=True, eq=True)
class Parenthesis(Expression):
    expr: Expression

    def __rich_console__(
        self, console: Console, options: ConsoleOptions
    ) -> RenderResult:
        yield Segment("(")
        yield self.expr
        yield Segment(")")

    def rich_highlights(self) -> Iterable[Span]:
        yield from self.expr.rich_highlights()


@dataclass(frozen=True, eq=True)
class ForTupleExpression(Expression):
    key_ident: Identifier | None
    value_ident: Identifier
    collection: Expression
    value: Expression
    condition: Expression | None

    def __rich_console__(
        self, console: Console, options: ConsoleOptions
    ) -> RenderResult:
        yield Segment("[")
        yield Segment("for ", style=STYLE_KEYWORDS)
        if self.key_ident is not None:
            yield self.key_ident
            yield Segment(", ")
        yield self.value_ident
        yield Segment(" in ", style=STYLE_KEYWORDS)
        yield self.collection
        yield Segment(" : ")
        yield self.value
        if self.condition is not None:
            yield Segment(" if ", style=STYLE_KEYWORDS)
            yield self.condition
        yield Segment("]")

    def rich_highlights(self) -> Iterable[Span]:
        yield SourceSpan(
            self.span.start + 1,
            self.key_ident.span.start - 1
            if self.key_ident is not None
            else self.value_ident.span.start - 1,
        ).styled(STYLE_KEYWORDS)

        if self.key_ident is not None:
            yield from self.key_ident.rich_highlights()
        yield from self.value_ident.rich_highlights()

        yield SourceSpan(
            self.value_ident.span.end + 1, self.collection.span.start - 1
        ).styled(STYLE_KEYWORDS)

        yield from self.collection.rich_highlights()
        yield from self.value.rich_highlights()
        if self.condition is not None:
            yield SourceSpan(
                self.value.span.end + 1, self.condition.span.start - 1
            ).styled(STYLE_KEYWORDS)
            yield from self.condition.rich_highlights()


@dataclass(frozen=True, eq=True)
class ForObjectExpression(Expression):
    key_ident: Identifier | None
    value_ident: Identifier
    collection: Expression
    key: Expression
    value: Expression
    condition: Expression | None
    grouping_mode: bool = dataclasses.field(default=False)

    def __rich_console__(
        self, console: Console, options: ConsoleOptions
    ) -> RenderResult:
        yield Segment("{")
        yield Segment("for ", style=STYLE_KEYWORDS)
        if self.key_ident is not None:
            yield self.key_ident
            yield Segment(", ")
        yield self.value_ident
        yield Segment(" in ", style=STYLE_KEYWORDS)
        yield self.collection
        yield Segment(" : ")
        yield self.key
        yield Segment(" => ")
        yield self.value
        if self.condition is not None:
            yield Segment(" if ", style=STYLE_KEYWORDS)
            yield self.condition
        if self.grouping_mode:
            yield Segment("...")
        yield Segment("}")

    def rich_highlights(self) -> Iterable[Span]:
        yield SourceSpan(
            self.span.start + 1,
            self.key_ident.span.start - 1
            if self.key_ident is not None
            else self.value_ident.span.start - 1,
        ).styled(STYLE_KEYWORDS)
        yield from (
            self.key_ident.rich_highlights() if self.key_ident is not None else []
        )
        yield from self.value_ident.rich_highlights()
        yield SourceSpan(
            self.value_ident.span.end + 1, self.collection.span.start - 1
        ).styled(STYLE_KEYWORDS)
        yield from self.collection.rich_highlights()
        yield from self.key.rich_highlights()
        if self.condition is not None:
            yield SourceSpan(
                self.value.span.end + 1, self.condition.span.start - 1
            ).styled(STYLE_KEYWORDS)
            yield from self.condition.rich_highlights()


class Stmt(Node):
    """Base class for nodes that represent statements in HCL2."""

    @property
    def key_path(self) -> tuple[str, ...]:
        raise NotImplementedError()


@dataclass(frozen=True, eq=True)
class Attribute(Stmt):
    key: Identifier
    value: Expression

    @property
    def key_path(self) -> tuple[str, ...]:
        return (self.key.name,)

    def __rich_console__(
        self, console: Console, options: ConsoleOptions
    ) -> RenderResult:
        yield Segment(self.key.name, style=STYLE_PROPERTY_NAME)
        yield Segment(" = ")
        yield self.value

    def rich_highlights(self) -> Iterable[Span]:
        yield self.key.span.styled(STYLE_PROPERTY_NAME)
        yield from self.value.rich_highlights()


@dataclass(frozen=True, eq=True)
class Block(Stmt):
    type: Identifier
    labels: list[Literal | Identifier]
    body: list[Stmt]

    @property
    def key_span(self) -> SourceSpan:
        return SourceSpan(
            self.type.span.start,
            self.labels[-1].span.end if self.labels else self.type.span.end,
        )

    @property
    def keys(self) -> tuple[String, ...]:
        key_parts: list[String] = [self.type.as_string()]
        for label in self.labels:
            match label:
                case Identifier() as label:
                    key_parts.append(label.as_string())
                case Literal(value=String() as string):
                    key_parts.append(string)
                case _:
                    pass
        return tuple(key_parts)

    @property
    def key_path(self) -> tuple[str, ...]:
        key_parts: list[str] = [self.type.name]
        for label in self.labels:
            if isinstance(label, Identifier):
                key_parts.append(label.name)
            elif isinstance(label, Literal) and isinstance(label.value, String):
                key_parts.append(label.value.raw())
        return tuple(key_parts)

    def key(self) -> tuple[str, ...]:
        return self.key_path

    def __rich_console__(
        self, console: Console, options: ConsoleOptions
    ) -> RenderResult:
        yield Segment(self.type.name, style=STYLE_KEYWORDS)
        for label in self.labels:
            yield Segment(" ")
            yield label
        yield Segment(" {")
        yield Segment("\n")
        for stmt in self.body:
            yield Padding(stmt, (0, 2))
        yield Segment("}")

    def rich_highlights(self) -> Iterable[Span]:
        yield self.type.span.styled(STYLE_KEYWORDS)
        for label in self.labels:
            yield from label.rich_highlights()
        for stmt in self.body:
            yield from stmt.rich_highlights()

    @cached_property
    def attributes(self) -> dict[str, Expression]:
        return {
            stmt.key.name: stmt.value
            for stmt in self.body
            if isinstance(stmt, Attribute)
        }

    @cached_property
    def blocks(self) -> list[Block]:
        return [stmt for stmt in self.body if isinstance(stmt, Block)]


@dataclass(frozen=True, eq=True)
class Module(Node):
    body: list[Stmt]

    def get_blocks(self, block_type: str | None) -> list[Block]:
        return [
            stmt
            for stmt in self.body
            if isinstance(stmt, Block)
            and (block_type is None or stmt.type.name == block_type)
        ]

    def get_block(self, block_type: str, *labels: str) -> Block | None:
        blocks = self.get_blocks(block_type)

        if len(labels) > 0:
            blocks = [
                block
                for block in blocks
                if block.labels == [Literal(String(label)) for label in labels]
            ]

        if len(blocks) > 1:
            raise ValueError(f"Multiple {block_type} blocks found")
        if len(blocks) == 0:
            return None
        return blocks[0]

    def __rich_console__(
        self, console: Console, options: ConsoleOptions
    ) -> RenderResult:
        for stmt in self.body:
            yield stmt
            yield Segment("\n")

    def rich_highlights(self) -> Iterable[Span]:
        for stmt in self.body:
            yield from stmt.rich_highlights()


class VarArgsMarker(Node):
    def __rich_console__(
        self, console: Console, options: ConsoleOptions
    ) -> RenderResult:
        yield Segment("...")



from __future__ import annotations

from typing import Any, TypeVar, cast

from pyagnostics.exceptions import DiagnosticError, DiagnosticErrorGroup
from pyagnostics.spans import LabeledSpan
from pydantic import BaseModel, ValidationError
from pydantic_core import ErrorDetails

from pyhcl2.eval import EvaluationScope, Evaluator
from pyhcl2.nodes import Block
from pyhcl2.rich_utils import Inline
from pyhcl2.values import Array, Object, String, Value

Model = TypeVar("Model", bound=BaseModel)


def load_model_from_block(
    block: Block,
    model_cls: type[Model],
    evaluator: Evaluator = Evaluator(),
    scope: EvaluationScope | None = None,
) -> Model:
    scope = scope or EvaluationScope()
    block_value: Object = cast(Object, evaluator.eval(block, scope))
    field_values: dict[str, Any] = {}
    field_context: dict[str, Value] = {}

    for k, v in block_value.items():
        name = k.raw()
        if name in model_cls.model_fields:
            field = model_cls.model_fields[name]

            field_context[name] = v
            if field.annotation is not None and (
                field.annotation is type(v) or field.annotation is Value
            ):
                field_values[name] = v
            else:
                field_values[name] = v.raw()

    try:
        return model_cls.model_validate(field_values, context=field_context)
    except ValidationError as e:
        diagnostics: list[DiagnosticError] = []
        for error in e.errors():
            error = cast(ErrorDetails, error)

            value: Value = block_value
            try:
                for loc in error["loc"]:
                    match loc:
                        case int():
                            assert isinstance(value, Array)
                            value = value[loc]
                        case str():
                            assert isinstance(value, Object)
                            value = value[String(loc)]
            except KeyError:
                pass

            field_key = list(block.key()) + [str(val) for val in error["loc"]]
            field_key_str = ".".join(field_key)

            match error["type"]:
                case "missing":
                    diagnostics.append(
                        DiagnosticError(
                            code="pyhcl2::models::validation_error",
                            message=f"Missing required field {field_key_str}",
                            labels=[
                                LabeledSpan(
                                    value.span, f"{value.type_name} missing field"
                                ),
                            ]
                            if value.span is not None
                            else [],
                        )
                    )
                case _:
                    diagnostics.append(
                        DiagnosticError(
                            code=f"pyhcl2::pydantic_validation_error::{error['type']}",
                            message=error["msg"],
                            labels=[
                                LabeledSpan(value.span, "invalid input"),
                            ]
                            if value.span is not None
                            else [],
                            notes=[Inline("[blue]context:[/blue] ", repr(error["ctx"]))]
                            if "ctx" in error
                            else [],
                        )
                    )

        raise DiagnosticErrorGroup("Failed to validate hcl model", diagnostics)
